import urllib.request
import logging
import mimetypes
from ..format import format_entry


logger = logging.getLogger(__name__)


class SiloCreationContext:
    def __init__(self, config, cache, silo_name):
        self.config = config
        self.cache = cache
        self.silo_name = silo_name


class SiloContextBase:
    def __init__(self, exec_ctx):
        self.exec_ctx = exec_ctx

    @property
    def args(self):
        return self.exec_ctx.args

    @property
    def config(self):
        return self.exec_ctx.config

    @property
    def cache(self):
        return self.exec_ctx.cache


class SiloAuthenticationContext(SiloContextBase):
    pass


class SiloPostingContext(SiloContextBase):
    pass


class Silo:
    SILO_TYPE = 'unknown'

    def __init__(self, ctx):
        self.ctx = ctx
        self._silo_cfg = dict(ctx.config.items('silo:%s' % self.name))

    @property
    def name(self):
        return self.ctx.silo_name

    def getConfigItem(self, name, fallback=None):
        return self._silo_cfg.get(name, fallback)

    def getConfigItems(self):
        return self._silo_cfg.copy()

    def getCacheItem(self, name, valtype=str):
        full_name = '%s_%s' % (self.name, name)
        return self.ctx.cache.getCustomValue(full_name, valtype=valtype)

    def setCacheItem(self, name, val):
        full_name = '%s_%s' % (self.name, name)
        return self.ctx.cache.setCustomValue(full_name, val)

    def formatEntry(self, entry, *args, **kwargs):
        return format_entry(entry, *args, **kwargs)

    def authenticate(self, ctx):
        raise NotImplementedError()

    def onPostStart(self, ctx):
        pass

    def postEntry(self, entry, ctx):
        raise NotImplementedError()

    def dryRunPostEntry(self, entry, ctx):
        pass

    def onPostEnd(self, ctx):
        pass


def _get_silo_section_names(config):
    return [sn for sn in config.sections() if sn.startswith('silo:')]


def has_any_silo(config):
    return bool(_get_silo_section_names(config))


def load_silos(config, cache):
    from .print import PrintSilo
    from .mastodon import MastodonSilo
    from .twitter import TwitterSilo
    from .webmention import WebmentionSilo
    silo_types = [PrintSilo, MastodonSilo, TwitterSilo, WebmentionSilo]
    silo_dict = dict([(s.SILO_TYPE, s) for s in silo_types])

    silos = []
    sec_names = _get_silo_section_names(config)
    for sec_name in sec_names:
        silo_name = sec_name[5:]
        sec_items = dict(config.items(sec_name))
        silo_type = sec_items.get('type')
        if not silo_type:
            raise Exception("No silo type specified for: %s" % silo_name)

        silo_class = silo_dict.get(silo_type)
        if not silo_class:
            raise Exception("Unknown silo type: %s" % silo_type)

        logger.debug("Creating silo '%s' for '%s'." % (silo_type, silo_name))
        cctx = SiloCreationContext(config, cache, silo_name)
        silo = silo_class(cctx)
        silos.append(silo)
    return silos


def upload_silo_media(entry, propname, callback):
    media_ids = None
    urls = entry.get(propname, [], force_list=True)
    if urls:
        media_ids = []
        for url in urls:
            mid = _do_upload_silo_media(url, callback)
            if mid is not None:
                media_ids.append(mid)
    return media_ids


def _do_upload_silo_media(url, callback):
    logger.debug("Downloading %s for upload to silo..." % url)
    mt, enc = mimetypes.guess_type(url)
    if not mt:
        mt = mimetypes.common_types['.jpg']

    ext = mimetypes.guess_extension(mt) or '.jpg'
    logger.debug("Got MIME type and extension: %s %s" % (mt, ext))

    try:
        tmpfile, headers = urllib.request.urlretrieve(url)
        logger.debug("Using temporary file: %s" % tmpfile)
        return callback(tmpfile, mt)
    finally:
        logger.debug("Cleaning up.")
        urllib.request.urlcleanup()



import io
import os
import os.path
import re
import logging
import tempfile
import pytest
import silorider.main


logger = logging.getLogger(__name__)


# def pytest_collect_file(parent, path):
#     if path.ext == ".html" and path.basename.startswith("feeds"):
#         return FeedFile(path, parent)


re_feed_test_sep = re.compile(r'^---$')


class FeedFile(pytest.File):
    def collect(self):
        with self.fspath.open() as fp:
            markup = fp.read()

        name = self.fspath.basename
        html_markup, yaml_markup = re_feed_test_sep.split(markup, 1)
        yield FeedItem(name, self, html_markup, yaml_markup)


class FeedItem(pytest.Item):
    def __init__(self, name, parent, in_spec, out_spec):
        super().__init__(name, parent)
        self.in_spec = in_spec
        self.out_spec = out_spec

    def runtest(self):
        pass


@pytest.fixture
def cli():
    return CliRunner()


class CliRunner:
    def __init__(self):
        self._cfgtxt = """
[cache]
uri=memory://for_test
"""
        self._feedcfg = []
        self._pre_hooks = []
        self._cleanup = []

    def getFeedPath(self, name):
        return os.path.join(os.path.dirname(__file__),
                            'feeds',
                            '%s.html' % name)

    def createTempFeed(self, contents):
        tmpfd, tmpname = tempfile.mkstemp()
        with os.fdopen(tmpfd, 'w', encoding='utf8') as tmpfp:
            tmpfp.write(contents)
        self._cleanup.append(tmpname)
        return tmpname

    def setConfig(self, cfgtxt):
        self._cfgtxt = cfgtxt
        return self

    def appendConfig(self, cfgtxt):
        self._cfgtxt += cfgtxt
        return self

    def appendFeedConfig(self, feed_name, feed_url):
        self._feedcfg.append((feed_name, feed_url))
        return self

    def setFeedConfig(self, feed_name, feed_url):
        self._feedcfg = []
        return self.appendFeedConfig(feed_name, feed_url)

    def appendSiloConfig(self, silo_name, silo_type, **options):
        cfgtxt = '[silo:%s]\n' % silo_name
        cfgtxt += 'type=%s\n' % silo_type
        if options is not None:
            for n, v in options.items():
                cfgtxt += '%s=%s\n' % (n, v)
        return self.appendConfig(cfgtxt)

    def preExecHook(self, hook):
        self._pre_hooks.append(hook)

    def run(self, *args):
        pre_args = ['-v']
        if self._cfgtxt or self._feedcfg:
            cfgtxt = self._cfgtxt
            cfgtxt += '\n\n[urls]\n'
            for n, u in self._feedcfg:
                cfgtxt += '%s=%s\n' % (n, u)

            tmpfd, tmpcfg = tempfile.mkstemp()
            logger.info("Creating temporary configuration file: %s" % tmpcfg)
            with os.fdopen(tmpfd, 'w') as tmpfp:
                tmpfp.write(cfgtxt)
            self._cleanup.append(tmpcfg)
            pre_args += ['-c', tmpcfg]

        captured = io.StringIO()
        handler = logging.StreamHandler(captured)
        handler.setLevel(logging.INFO)
        silorider_logger = logging.getLogger('silorider')
        silorider_logger.addHandler(handler)

        main_ctx = None
        main_res = None

        def pre_exec_hook(ctx):
            for h in self._pre_hooks:
                h(ctx)

        def post_exec_hook(ctx, res):
            nonlocal main_ctx, main_res
            main_ctx = ctx
            main_res = res

        silorider.main.pre_exec_hook = pre_exec_hook
        silorider.main.post_exec_hook = post_exec_hook

        args = pre_args + list(args)
        logger.info("Running command: %s" % list(args))
        try:
            silorider.main._unsafe_main(args)
        finally:
            silorider.main.pre_exec_hook = None
            silorider.main.post_exec_hook = None

            silorider_logger.removeHandler(handler)

            logger.info("Cleaning %d temporary files." % len(self._cleanup))
            for tmpname in self._cleanup:
                try:
                    os.remove(tmpname)
                except FileNotFoundError:
                    pass

        return main_ctx, main_res


@pytest.fixture
def feedutil():
    return FeedUtil()


class FeedUtil:
    def makeFeed(self, *entries):
        feed = '<html><body>\n'
        for e in entries:
            feed += '<article class="h-entry">\n'
            feed += e
            feed += '</article>\n'
        feed += '</body></html>'
        return feed



import logging
import twitter
from .base import Silo


logger = logging.getLogger(__name__)


class TwitterSilo(Silo):
    SILO_TYPE = 'twitter'
    _CLIENT_CLASS = twitter.Api

    def __init__(self, ctx):
        super().__init__(ctx)
        self.client = None

    def authenticate(self, ctx):
        force = ctx.exec_ctx.args.force

        client_token = self.getCacheItem('clienttoken')
        if not client_token or force:
            logger.info("Please enter Twitter consumer tokens for %s:" %
                        self.ctx.silo_name)
            consumer_key = input("Consumer Key: ")
            consumer_secret = input("Consumer Secret: ")
            client_token = '%s,%s' % (consumer_key, consumer_secret)
            self.setCacheItem('clienttoken', client_token)

        access_token = self.getCacheItem('accesstoken')
        if not access_token or force:
            logger.info("Please enter Twitter access tokens for %s:" %
                        self.ctx.silo_name)

            access_key = input("Access Token: ")
            access_secret = input("Access Token Secret: ")

            access_token = '%s,%s' % (access_key, access_secret)
            self.setCacheItem('accesstoken', access_token)

    def onPostStart(self, ctx):
        if not ctx.args.dry_run:
            self._ensureClient()

    def _ensureClient(self):
        if self.client is not None:
            return

        logger.debug("Creating Twitter API client.")
        client_token = self.getCacheItem('clienttoken')
        if not client_token:
            raise Exception("Twitter silo '%s' isn't authenticated." %
                            self.name)

        client_key, client_secret = client_token.split(',')

        access_token = self.getCacheItem('accesstoken')
        if not access_token:
            raise Exception("Twitter silo '%s' isn't authenticated." %
                            self.name)

        access_key, access_secret = access_token.split(',')

        self.client = self._CLIENT_CLASS(
            consumer_key=client_key,
            consumer_secret=client_secret,
            access_token_key=access_key,
            access_token_secret=access_secret)

    def postEntry(self, entry, ctx):
        tweettxt = self.formatEntry(entry, limit=280,
                                    url_flattener=TwitterUrlFlattener())
        if not tweettxt:
            raise Exception("Can't find any content to use for the tweet!")

        logger.debug("Posting tweet: %s" % tweettxt)
        media_urls = entry.get('photo', [], force_list=True)
        self.client.PostUpdate(tweettxt, media=media_urls)

    def dryRunPostEntry(self, entry, ctx):
        tweettxt = self.formatEntry(entry, limit=280)
        logger.info("Tweet would be:")
        logger.info(tweettxt)


TWITTER_NETLOCS = ['twitter.com', 'www.twitter.com']


class TwitterUrlFlattener:
    def replaceHref(self, text, url):
        # Is it a Twitter URL?
        if url.netloc not in TWITTER_NETLOCS:
            return None

        path = url.path.lstrip('/')
        # Is it a profile URL?
        if '/' not in path:
            return '@' + path

        return None



import logging
from ..silos.base import SiloAuthenticationContext


logger = logging.getLogger(__name__)


def auth_silo(ctx):
    silo_names = ctx.args.silo
    if 'all' in silo_names:
        silo_names = [s.name for s in ctx.silos]

    for silo in ctx.silos:
        if silo.name not in silo_names:
            continue

        logger.debug("Authenticating silo: %s" % silo.name)
        authctx = SiloAuthenticationContext(ctx)
        silo.authenticate(authctx)



import logging
import ronkyuu
from .base import Silo
from ..config import has_lxml


logger = logging.getLogger(__name__)


class WebmentionSilo(Silo):
    SILO_TYPE = 'webmention'

    def __init__(self, ctx):
        super().__init__(ctx)
        self.client = None
        ronkyuu.setParser('lxml' if has_lxml else 'html5lib')

    def authenticate(self, ctx):
        logger.info("Webmention silo doesn't require authentication.")

    def postEntry(self, entry, ctx):
        source_url = entry.url
        logger.debug("Finding mentions in: %s" % source_url)
        refs = ronkyuu.findMentions(source_url)
        for r in refs.get('refs', []):
            logger.debug("Sending webmention: %s -> %s" % (source_url, r))
            ronkyuu.sendWebmention(source_url, r)



import os
import os.path
import logging
import argparse
import configparser
import coloredlogs


logger = logging.getLogger(__name__)


class ExecutionContext:
    def __init__(self, args, config, cache, silos):
        self.args = args
        self.config = config
        self.cache = cache
        self.silos = silos


def _setup_auth(parser):
    def _run(ctx):
        from .commands.auth import auth_silo
        auth_silo(ctx)

    parser.add_argument(
        'silo',
        action='append',
        help=("The name of the silo to authenticate. "
              "Use 'all' to authenticate all declared silos."))
    parser.add_argument(
        '-f', '--force',
        action='store_true',
        help="Force re-authenticate even silos with valid access.")
    parser.add_argument(
        '--console',
        action='store_true',
        help=("Use the current terminal to enter credentials. This is "
              "useful if you're not in an environment where silorider can "
              "launch a browser."))
    parser.set_defaults(func=_run)


def _setup_process(parser):
    def _run(ctx):
        from .commands.process import process_urls
        process_urls(ctx)

    parser.add_argument(
        '-u', '--url',
        action='append',
        help="Only parse the given URL name(s).")
    parser.add_argument(
        '-s', '--silo',
        action='append',
        help="Only use the given silo(s).")
    parser.add_argument(
        '--no-cache',
        action='store_true',
        help="Ignore the cache, post all entries that qualify.")
    parser.add_argument(
        '--since',
        help="Post entries since the specified date/time only.")
    parser.add_argument(
        '--dry-run',
        action='store_true',
        help="Only report what would be posted, but don't post anything.")
    parser.set_defaults(func=_run)


def _setup_populate(parser):
    def _run(ctx):
        from .commands.utils import populate_cache
        populate_cache(ctx)

    parser.add_argument(
        '-u', '--url',
        action='append',
        help="Only populate from the given URL name(s).")
    parser.add_argument(
        '-s', '--silo',
        action='append',
        help="Only populate the given silo(s).")
    parser.add_argument(
        '--until',
        help="The date until which to populate the cache (included).")
    parser.set_defaults(func=_run)


commands = {
    'auth': {
        'help': "Authenticate with a silo service.",
        'setup': _setup_auth,
    },
    'process': {
        'help': "Post a website's latest articles to silo services.",
        'setup': _setup_process,
    },
    'populate': {
        'help': "Populates the cache with the latest entries from a feed.",
        'setup': _setup_populate,
    }
}


has_debug_logging = False
pre_exec_hook = None
post_exec_hook = None


def _unsafe_main(args=None):
    parser = argparse.ArgumentParser('SiloRider')
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        help="Print debug messages.")
    parser.add_argument(
        '--no-color',
        action='store_true',
        help="Don't use pretty colors.")
    parser.add_argument(
        '-c', '--config',
        help="Configuration file to load.")

    subparsers = parser.add_subparsers()
    for cn, cd in commands.items():
        cp = subparsers.add_parser(cn, help=cd.get('help'))
        cd['setup'](cp)

    args = parser.parse_args(args)

    global has_debug_logging
    has_debug_logging = args.verbose

    if not args.no_color:
        coloredlogs.install()

    loglvl = logging.DEBUG if args.verbose else logging.INFO
    root_logger = logging.getLogger()
    root_logger.setLevel(loglvl)
    for handler in root_logger.handlers:
        handler.setLevel(loglvl)

    if not getattr(args, 'func', None):
        parser.print_help()
        return

    logger.debug("Loading configuration.")
    xdg_config_home = os.getenv('XDG_CONFIG_HOME',
                                os.path.expanduser('~/.config'))
    config = configparser.ConfigParser(interpolation=None)
    config_paths = [
        os.path.join(os.path.dirname(__file__), 'default.cfg'),
        os.path.join(xdg_config_home, 'silorider/silorider.cfg')
    ]
    if args.config:
        config_paths.append(args.config)
    config.read(config_paths)

    from .silos.base import has_any_silo
    if not has_any_silo(config):
        logger.warning("No silos defined in the configuration file. "
                       "Nothing to do!")
        return
    if not config.has_section('urls') or not config.items('urls'):
        logger.warning("No URLs defined in the configuration file. "
                       "Nothing to do!")
        return

    logger.debug("Initializing cache.")
    from .cache.base import load_cache
    cfg_dir = os.path.dirname(args.config) if args.config else None
    cache = load_cache(config, cfg_dir)

    logger.debug("Initializing silo riders.")
    from .silos.base import load_silos
    silos = load_silos(config, cache)

    ctx = ExecutionContext(args, config, cache, silos)

    if pre_exec_hook:
        pre_exec_hook(ctx)

    res = args.func(ctx)

    if post_exec_hook:
        post_exec_hook(ctx, res)

    if isinstance(res, int):
        return res
    return 0


def main():
    try:
        res = _unsafe_main()
    except Exception as ex:
        if has_debug_logging:
            raise
        logger.error(ex)
        res = 1

    import sys
    sys.exit(res)


if __name__ == '__main__':
    main()



import unittest.mock
import requests


def test_one_article_no_mentions(cli, feedutil):
    feed = cli.createTempFeed(feedutil.makeFeed(
        """<h1 class="p-name">A new article</h1>
<div class="e-content">
<p>This is the abstract of the article.</p>
<p>Read more at <a class="u-url" href="https://example.org/a-new-article">permalink</a>.</p>
</div>
"""  # NOQA
    ))

    cli.appendSiloConfig('test', 'webmention', url='/blah')
    cli.setFeedConfig('feed', feed)

    with unittest.mock.patch('requests.get') as mock_get, \
            unittest.mock.patch('requests.post') as mock_post:
        mock_get.side_effect = [
            _MockResponse('')]
        mock_post.side_effect = []
        ctx, _ = cli.run('process')
        assert mock_get.call_args_list[0][0] == ('https://example.org/a-new-article',)  # NOQA


def test_one_article_one_mention(cli, feedutil):
    feed = cli.createTempFeed(feedutil.makeFeed(
        """<h1 class="p-name">A new article</h1>
<div class="e-content">
<p>This is the abstract of the article.</p>
<p>Read more at <a class="u-url" href="https://example.org/a-new-article">permalink</a>.</p>
</div>
"""  # NOQA
    ))

    cli.appendSiloConfig('test', 'webmention', url='/blah')
    cli.setFeedConfig('feed', feed)

    with unittest.mock.patch('requests.get') as mock_get, \
            unittest.mock.patch('requests.post') as mock_post:
        mock_get.side_effect = [
            _MockResponse("""
<p>This is a reply to <a href="https://other.org/article">another article<a>.</p>
"""),  # NOQA
            _MockResponse("""
<html><head>
    <link rel="webmention" href="https://other.org/webmention">
</head><body>
</body></html>""")]
        mock_post.side_effect = [
            _MockResponse('')]
        ctx, _ = cli.run('process')
        assert mock_get.call_args_list[0][0] == ('https://example.org/a-new-article',)  # NOQA
        assert mock_get.call_args_list[1][0] == ('https://other.org/article',)  # NOQA
        assert mock_post.call_args_list[0][0] == ('https://other.org/webmention',)  # NOQA


class _MockResponse:
    def __init__(self, txt):
        self.status_code = requests.codes.ok
        self.headers = {}
        self.history = []
        self.text = self.content = txt



import pytest
from silorider.format import format_entry, strip_html


test_url = 'https://example.org/article'


def _make_test_entry(best_name, is_micropost):
    class TestEntry:
        def __init__(self):
            self.is_micropost = is_micropost
            self.url = test_url

        def get(self, _):
            return best_name

        def htmlFind(self, *args, **kwargs):
            return best_name

    entry = TestEntry()
    return entry


@pytest.mark.parametrize("text, expected", [
    ("<p>Something</p>",
     "Something"),
    ("<p>Something with <em>emphasis</em> in it</p>",
     "Something with emphasis in it"),
    ("<p>Something with <a href=\"http://example.org/blah\">a link</a>",
     "Something with a link http://example.org/blah"),
    ("<p>Something with a link <a href=\"http://example.org/blah\">http://example.org</a>",  # NOQA
     "Something with a link http://example.org/blah"),
    ("<p>Something with <a href=\"http://example.org/first\">one link here</a> and <a href=\"http://example.org/second\">another there</a>...</p>",  # NOQA
     "Something with one link here http://example.org/first and another there http://example.org/second...")  # NOQA
])
def test_strip_html(text, expected):
    actual = strip_html(text)
    assert actual == expected


@pytest.mark.parametrize("text, expected", [
    ("<p>Something with <a href=\"http://example.org/blah\">a link</a>",
     "Something with a link\nhttp://example.org/blah"),
    ("<p>Something with a link <a href=\"http://example.org/blah\">http://example.org</a>",  # NOQA
     "Something with a link http://example.org/blah"),
    ("<p>Something with <a href=\"http://example.org/first\">one link here</a> and <a href=\"http://example.org/second\">another there</a>...</p>",  # NOQA
     "Something with one link here and another there...\nhttp://example.org/first\nhttp://example.org/second")  # NOQA
])
def test_strip_html_with_bottom_urls(text, expected):
    actual = strip_html(text, inline_urls=False)
    print(actual)
    print(expected)
    assert actual == expected


@pytest.mark.parametrize("title, limit, add_url, expected", [
    ('A test entry', None, False, 'A test entry'),
    ('A test entry', None, 'auto', 'A test entry ' + test_url),
    ('A test entry', None, True, 'A test entry ' + test_url),

    ('A test entry', 80, False, 'A test entry'),
    ('A test entry', 80, 'auto', 'A test entry ' + test_url),
    ('A test entry', 80, True, 'A test entry ' + test_url),

    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, False,
     'A test entry that is very very long because its title has many many '
     'words in...'),
    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, 'auto',
     'A test entry that is very very long because its... ' + test_url),
    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, True,
     'A test entry that is very very long because its... ' + test_url)
])
def test_format_lonform_entry(title, limit, add_url, expected):
    entry = _make_test_entry(title, False)
    actual = format_entry(entry, limit, add_url)
    assert actual == expected


@pytest.mark.parametrize("text, limit, add_url, expected", [
    ('A test entry', None, False, 'A test entry'),
    ('A test entry', None, 'auto', 'A test entry'),
    ('A test entry', None, True, 'A test entry ' + test_url),

    ('A test entry', 80, False, 'A test entry'),
    ('A test entry', 80, 'auto', 'A test entry'),
    ('A test entry', 80, True, 'A test entry ' + test_url),

    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, False,
     'A test entry that is very very long because its title has many many '
     'words in...'),
    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, 'auto',
     'A test entry that is very very long because its... ' + test_url),
    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, True,
     'A test entry that is very very long because its... ' + test_url)
])
def test_format_micropost_entry(text, limit, add_url, expected):
    entry = _make_test_entry(text, True)
    actual = format_entry(entry, limit, add_url)
    assert actual == expected



import re
import urllib.parse
import textwrap
import bs4
from .config import has_lxml


def format_entry(entry, limit=None, add_url='auto', url_flattener=None):
    url = entry.url
    name = get_best_text(entry, url_flattener=url_flattener)
    if not name:
        raise Exception("Can't find best text for entry: %s" % url)

    do_add_url = ((add_url is True) or
                  (add_url == 'auto' and not entry.is_micropost))
    if limit:
        if do_add_url and url:
            limit -= 1 + len(url)

        shortened = len(name) > limit
        if shortened:
            # If we have to shorten the text, but we haven't taken the
            # URL into account yet, let's see if we have to include now!
            # (this happens when we only want to include it when the text
            #  is shortened)
            if not do_add_url and add_url == 'auto' and url:
                do_add_url = True
                limit -= 1 + len(url)

        if limit <= 0:
            raise Exception("Can't shorten post name.")

        name = textwrap.shorten(name, width=limit, placeholder="...")

    if do_add_url and url:
        name += ' ' + url
    return name


class UrlFlattener:
    def replaceHref(self, text, url):
        return None


def get_best_text(entry, *, plain=True, inline_urls=True, url_flattener=None):
    elem = entry.htmlFind(class_='p-title')
    if not elem:
        elem = entry.htmlFind(class_='p-name')
    if not elem:
        elem = entry.htmlFind(class_='e-content')

    if elem:
        if not plain:
            text = '\n'.join([str(c) for c in elem.contents])
            return str(text)
        return strip_html(elem, inline_urls=inline_urls,
                          url_flattener=url_flattener)

    return None


def strip_html(bs_elem, *, inline_urls=True, url_flattener=None):
    if isinstance(bs_elem, str):
        bs_elem = bs4.BeautifulSoup(bs_elem,
                                    'lxml' if has_lxml else 'html5lib')

    outtxt = ''
    ctx = _HtmlStripping()
    ctx.url_flattener = url_flattener
    for c in bs_elem.children:
        outtxt += _do_strip_html(c, ctx)

    keys = ['url:%d' % i for i in range(len(ctx.urls))]
    if inline_urls:
        urls = dict(zip(keys, [' ' + u for u in ctx.urls]))
    else:
        urls = dict(zip(keys, [''] * len(ctx.urls)))
    outtxt = outtxt % urls
    if not inline_urls and ctx.urls:
        outtxt += '\n' + '\n'.join(ctx.urls)
    return outtxt


class _HtmlStripping:
    def __init__(self):
        self.urls = []
        self.url_flattener = None


def _escape_percents(txt):
    return txt.replace('%', '%%')


def _do_strip_html(elem, ctx):
    if isinstance(elem, bs4.NavigableString):
        return _escape_percents(str(elem))

    if elem.name == 'a':
        try:
            href = elem['href']
        except KeyError:
            href = None
        cnts = list(elem.contents)
        if len(cnts) == 1:
            href_txt = cnts[0].string
            if href_txt in href:
                # If we have a simple hyperlink where the text is a
                # substring of the target URL, just return the URL.
                return _escape_percents(href)

            if ctx.url_flattener:
                # Use an URL flattener if we have one.
                href_parsed = urllib.parse.urlparse(href)
                href_flattened = ctx.url_flattener.replaceHref(
                    href_txt, href_parsed)
                if href_flattened is not None:
                    return href_flattened

        # No easy way to simplify this hyperlink... let's put a marker
        # for the URL to be later replaced in the text.
        a_txt = ''.join([_do_strip_html(c, ctx)
                         for c in cnts])
        a_txt += '%%(url:%d)s' % len(ctx.urls)
        ctx.urls.append(href)
        return a_txt

    if elem.name == 'ol':
        outtxt = ''
        for i, c in enumerate(elem.children):
            if c.name == 'li':
                outtxt += ('%s. ' % (i + 1)) + _do_strip_html(c, ctx)
                outtxt += '\n'
        return outtxt

    if elem.name == 'ul':
        outtxt = ''
        for c in elem.children:
            if c.name == 'li':
                outtxt += '- ' + _do_strip_html(c, ctx)
                outtxt += '\n'
        return outtxt

    return ''.join([_do_strip_html(c, ctx) for c in elem.children])


re_sentence_end = re.compile(r'[\w\]\)\"\'\.]\.\s|[\?\!]\s')


def shorten_text(txt, limit):
    if len(txt) <= limit:
        return (txt, False)

    m = re_sentence_end.search(txt)
    if m and m.end <= (limit + 1):
        return (txt[:m.end - 1], True)

    shorter = textwrap.shorten(
        txt, width=limit, placeholder="...")
    return (shorter, True)



from .base import Cache


class MemoryCache(Cache):
    def __init__(self):
        self._vals = {}
        self._posted = {}

    def getCustomValue(self, name, valtype=str):
        return self._vals.get(name)

    def setCustomValue(self, name, val):
        self._vals[name] = val

    def wasPosted(self, silo_name, entry_uri):
        uris = self._posted.get(silo_name)
        if uris:
            return entry_uri in uris
        return False

    def addPost(self, silo_name, entry_uri):
        uris = self._posted.setdefault(silo_name, set())
        uris.add(entry_uri)



import logging
from .utils import get_named_silos, get_named_urls
from ..silos.base import SiloPostingContext
from ..parse import parse_url


logger = logging.getLogger(__name__)


def process_urls(ctx):
    for url in get_named_urls(ctx.config, ctx.args.url):
        logger.info("Processing %s" % url)
        p = Processor(ctx, url)
        p.process()


class Processor:
    def __init__(self, ctx, url):
        self.ctx = ctx
        self.url = url
        self._silos = get_named_silos(ctx.silos, ctx.args.silo)

    @property
    def config(self):
        return self.ctx.config

    @property
    def silos(self):
        return self._silos

    def process(self):
        self.preProcess()

        feed = parse_url(self.url)
        for entry in feed.entries:
            self.processEntry(entry)

        self.postProcess()

    def preProcess(self):
        # Pre-parse the "since" date/time.
        if self.ctx.args.since:
            import dateparser
            self.ctx.args.since = dateparser.parse(self.ctx.args.since)

        for silo in self.silos:
            silo.onPostStart(self.ctx)

    def postProcess(self):
        for silo in self.silos:
            silo.onPostEnd(self.ctx)

    def processEntry(self, entry):
        entry_url = entry.get('url')
        if not entry_url:
            logger.warning("Found entry without a URL.")
            return

        if self.isEntryFiltered(entry):
            logger.debug("Entry is filtered out: %s" % entry_url)
            return

        postctx = SiloPostingContext(self.ctx)
        no_cache = self.ctx.args.no_cache
        only_since = self.ctx.args.since
        logger.debug("Processing entry: %s" % entry_url)
        for silo in self.silos:
            if only_since:
                entry_dt = entry.get('published')
                if not entry_dt:
                    logger.warning(
                        "Skipping entry with no published date/time "
                        "for %s: %s" % (silo.name, entry_url))
                    continue

                # Strip entry datetime's time-zone information if we
                # don't have a time-zone info from the command line.
                if not only_since.tzinfo:
                    entry_dt = entry_dt.replace(tzinfo=None)

                if entry_dt < only_since:
                    logger.info(
                        "Skipping entry older than specified date/time "
                        "for %s: %s" % (silo.name, entry_url))
                    continue

            if not no_cache and self.ctx.cache.wasPosted(silo.name, entry_url):
                logger.debug("Skipping already posted entry on %s: %s" %
                             (silo.name, entry_url))
                continue

            if not self.ctx.args.dry_run:
                try:
                    did_post = silo.postEntry(entry, postctx)
                except Exception as ex:
                    did_post = False
                    logger.error("Error posting: %s" % entry_url)
                    logger.error(ex)
                    if self.ctx.args.verbose:
                        raise
                if did_post is True or did_post is None:
                    self.ctx.cache.addPost(silo.name, entry_url)
            else:
                logger.info("Would post entry on %s: %s" %
                            (silo.name, entry_url))
                silo.dryRunPostEntry(entry, postctx)

    def isEntryFiltered(self, entry):
        if not self.config.has_section('filter'):
            return False

        items = self.config.items('filter')

        for name, value in items:
            if name.startswith('include_'):
                propname = name[8:]
                propvalue = entry.get(propname)
                for inc_val in value.split(','):
                    if inc_val in propvalue:
                        break
                else:
                    return True

            elif name.startswith('exclude_'):
                propname = name[8:]
                propvalue = entry.get(propname)
                for excl_val in value.split(','):
                    if excl_val in propvalue:
                        return True

        return False
